from collections import OrderedDict, namedtuple
import json
import sys
import time

from contracts import contract
from contracts.utils import indent
from system_cmd import CmdException, system_cmd_result

from .utils import memoize_simple, tmpfile


class InvalidHash(Exception):
    pass


IPFS_ls_entry = namedtuple('IPFS_ls_entry', 'name hash size')


class IPFSInterface():

    def __init__(self):
        self.providers_ttl = 120
        self.providers_cache = {}
        self.debug = False

    def _cmd(self, cmd):

        if self.debug:
            print('$ %s' % " ".join(cmd))
        else:
            sys.stderr.write('.')
        t0 = time.time()
        res = system_cmd_result('.', cmd, raise_on_error=True)

        delta = time.time() - t0
        if self.debug:
            print('took %5.2f s: $ %s' % (delta, " ".join(cmd)))
        return res

    def get_keys(self):
        cmd = ['ipfs', 'key', 'list', '-l']
        res = self._cmd(cmd)
        #print res.stdout.__repr__()
        lines = res.stdout.strip().split('\n')
        res = OrderedDict()
        for l in lines:
            tokens = l.strip().split(' ')
            assert len(tokens) == 2, tokens
            res[tokens[1]] = tokens[0]
        return res

    def gen_key(self, name, key_type, key_size):
        cmd = ['ipfs', 'key', 'gen', '--type', key_type, '--size', str(key_size), name]
        res = self._cmd(cmd)
        ipfs_hash = res.stdout.strip()
        return ipfs_hash

    def publish(self, key_name, ipfs_hash, timeout='60s'):
        cmd = ['ipfs', 'name', 'publish', '--timeout', timeout, '--key', key_name, ipfs_hash]
        _res = self._cmd(cmd)

    def pin_ls(self):
        cmd = ['ipfs', 'pin', 'ls']
        res = self._cmd(cmd)
        recursive = set()
        indirect = set()
        for line in res.stdout.strip().split('\n'):
            tokens = line.split(' ')
            hashed = tokens[0]
            if tokens[1] == 'recursive':
                recursive.add(hashed)
            elif tokens[1] == 'indirect':
                indirect.add(hashed)
            else:
                assert False, line

        return recursive, indirect

    def dht_findprovs(self, ipfs_hash, timeout="1s"):
        if ipfs_hash in self.providers_cache:
            t, result = self.providers_cache[ipfs_hash]
            elapsed = time.time() - t
            if elapsed < self.providers_ttl:
                return result
        result = self._dht_findprovs(ipfs_hash, timeout)
        self.providers_cache[ipfs_hash] = time.time(), result
        return result

    def _dht_findprovs(self, ipfs_hash, timeout):
        cmd = ['ipfs', 'dht', 'findprovs', '--timeout', timeout, ipfs_hash]
        try:
            res = self._cmd(cmd)
        except CmdException as e:
            res = e.res
        options = res.stdout.strip().split('\n')
        options = [x for x in options if x]
        return options

    @memoize_simple
    def object_get(self, h, timeout):
        cmd = ['ipfs', 'object', '--timeout', timeout, 'get', h]
        try:
            res = self._cmd(cmd)
        except CmdException as e:
            raise InvalidHash(e.res.stderr)
        return res.stdout

    @memoize_simple
    def get(self, h, timeout="1h"):
        with tmpfile('.ipfs_data') as f:
            cmd = ['ipfs', 'get', '--timeout', timeout, '-o', f, h]
            _res = self._cmd(cmd)
            data = open(f).read()
            return data

    def ls(self, h, timeout="10s"):
    #    {"Links":[{"Name":"FILE2","Hash":"QmUtkGLvPf63NwVzLPKPUYgwhn8ZYPWF6vKWN3fZ2amfJF","Size":14},
    # {"Name":"upload_info1.yaml","Hash":"QmeiuS7VWRaUQTmva3UMgggCCEDisgtLWPwxFnJ1kCGaDJ","Size":214}],"Data":"\u0008\u0001"}
        data = self.object_get(h, timeout=timeout)
        d = json.loads(data)
        entries = OrderedDict()
        for entry in d['Links']:
            entries[str(entry['Name'])] = \
                IPFS_ls_entry(str(entry['Name']), str(entry['Hash']), entry['Size'])
        return entries

    def add_bytes(self, s):
        return get_hash_for_bytes_(s)

    def block_put(self, s):
        cmd = ['ipfs', 'block', 'put']
        res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True, display_stdout=False, display_stderr=False,
                                write_stdin=s)
        token = res.stdout.strip()
        return token

    def _get_ipfs_id(self):
        cmd = ['ipfs', 'id']
        res = self._cmd(cmd)
        data = json.loads(res.stdout.strip())
        return data

    @memoize_simple
    def ipfs_id(self):
        data = self._get_ipfs_id()
        return str(data['ID'])

    @memoize_simple
    @contract(returns='list(str)')
    def get_addresses(self):
        data = self._get_ipfs_id()
        return map(str, data['Addresses'])


def get_hash_for_bytes_(s):
    cmd = ['ipfs', 'add']
    cwd = '.'
    res = system_cmd_result(cwd, cmd, raise_on_error=True,
                            write_stdin=s)
    hashed = res.stdout.split()[1]
    assert 'Qm' in hashed, hashed
    return hashed


def add_ipfs_dir(dirname):
    cmd = ['ipfs', 'add', '-r', dirname]
    res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True)
    s = res.stdout.strip()
    lines = s.split('\n')
    out = lines[-1].split(' ')
    if (len(out) < 3 or out[0] != 'added' or not out[1].startswith('Qm')):
        msg = 'Invalid output for ipds:\n%s' % indent(res.stdout, ' > ')
        raise Exception(msg)
    hashed = out[1]
    return hashed

#
#def _get_ipfs_id():
#    cmd = ['ipfs', 'id']
#    res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True,
#                             display_stdout=False, display_stderr=False)
#    data = json.loads(res.stdout.strip())
#    return str(data['ID'])


class MakeIPFS2(object):

    def __init__(self):
        self.links = []
        self.total_file_size = 0

    def add_file(self, filename, ipfs, size):
        x = {'Name': filename, 'Hash': ipfs, "Size": size}
        self.links.append(x)
        self.total_file_size += size

    def add_file_content(self, filename, s):
        hashed = get_hash_for_bytes_(s)
        self.add_file(filename, hashed, len(s))

    def get_dag(self):
        result = {'Data': u"\u0008\u0001", 'Links': self.links}
        return result

    def total_size(self):
        return self.total_file_size

    def get_ipfs_hash(self):
        dag = self.get_dag()
        dag_json = json.dumps(dag)

        cmd = ['ipfs', 'object', 'put']
        cwd = '.'
        res = system_cmd_result(cwd, cmd, raise_on_error=True,
                                write_stdin=dag_json)
        hashed = res.stdout.split()[1]
        assert 'Qm' in hashed, hashed
        return hashed

#
#
#def ipfs_pubsub_send(topic, message):
#    cmd = ['ipfs', 'pubsub', 'pub', topic, message + '\n']
#    system_cmd_result('.', cmd, raise_on_error=True)
